import Foundation

public struct Gateway: Sendable, Codable {
    
    /// https://discord.com/developers/docs/topics/opcodes-and-status-codes#opcodes-and-status-codes
    public enum Opcode: UInt8, Sendable, Codable, CustomStringConvertible {
        case dispatch = 0
        case heartbeat = 1
        case identify = 2
        case presenceUpdate = 3
        case voiceStateUpdate = 4
        case resume = 6
        case reconnect = 7
        case requestGuildMembers = 8
        case invalidSession = 9
        case hello = 10
        case heartbeatAccepted = 11
        
        public var description: String {
            switch self {
            case .dispatch: return "dispatch"
            case .heartbeat: return "heartbeat"
            case .identify: return "identify"
            case .presenceUpdate: return "presenceUpdate"
            case .voiceStateUpdate: return "voiceStateUpdate"
            case .resume: return "resume"
            case .reconnect: return "reconnect"
            case .requestGuildMembers: return "requestGuildMembers"
            case .invalidSession: return "invalidSession"
            case .hello: return "hello"
            case .heartbeatAccepted: return "heartbeatAccepted"
            }
        }
    }
    
    /// The top-level gateway event.
    /// https://discord.com/developers/docs/topics/gateway#gateway-events
    public struct Event: Sendable, Codable {
        
        /// This enum is just for swiftly organizing Discord gateway event's `data`.
        /// You need to read each case's inner payload's documentation for more info.
        public enum Payload: Sendable {
            /// https://discord.com/developers/docs/topics/gateway-events#heartbeat
            case heartbeat(lastSequenceNumber: Int?)
            case identify(Identify)
            case hello(Hello)
            case ready(Ready)
            /// Is sent when we want to send a resume request
            case resume(Resume)
            /// Is received when Discord has ended replying our lost events, after a resume
            /// https://discord.com/developers/docs/topics/gateway-events#resumed
            case resumed
            /// https://discord.com/developers/docs/topics/gateway-events#invalid-session
            case invalidSession(canResume: Bool)
            
            case channelCreate(DiscordChannel)
            case channelUpdate(DiscordChannel)
            case channelDelete(DiscordChannel)
            case channelPinsUpdate(ChannelPinsUpdate)
            
            case threadCreate(DiscordChannel)
            case threadUpdate(DiscordChannel)
            case threadDelete(ThreadDelete)
            
            case threadSyncList(ThreadListSync)
            case threadMemberUpdate(ThreadMemberUpdate)
            case threadMembersUpdate(ThreadMembersUpdate)
            
            case guildCreate(GuildCreate)
            case guildUpdate(Guild)
            case guildDelete(UnavailableGuild)
            
            case guildBanAdd(GuildBan)
            case guildBanRemove(GuildBan)
            
            case guildEmojisUpdate(GuildEmojisUpdate)
            case guildStickersUpdate(GuildStickersUpdate)
            case guildIntegrationsUpdate(GuildIntegrationsUpdate)
            
            case guildMemberAdd(GuildMemberAdd)
            case guildMemberRemove(GuildMemberRemove)
            case guildMemberUpdate(GuildMemberAdd)
            
            case guildMembersChunk(GuildMembersChunk)
            case requestGuildMembers(RequestGuildMembers)
            
            case guildRoleCreate(GuildRole)
            case guildRoleUpdate(GuildRole)
            case guildRoleDelete(GuildRoleDelete)
            
            case guildScheduledEventCreate(GuildScheduledEvent)
            case guildScheduledEventUpdate(GuildScheduledEvent)
            case guildScheduledEventDelete(GuildScheduledEvent)
            
            case guildScheduledEventUserAdd(GuildScheduledEventUser)
            case guildScheduledEventUserRemove(GuildScheduledEventUser)
            
            case guildAuditLogEntryCreate(AuditLog.Entry)
            
            case integrationCreate(IntegrationCreate)
            case integrationUpdate(IntegrationCreate)
            case integrationDelete(IntegrationDelete)
            
            case interactionCreate(Interaction)
            
            case inviteCreate(InviteCreate)
            case inviteDelete(InviteDelete)
            
            case messageCreate(MessageCreate)
            case messageUpdate(DiscordChannel.PartialMessage)
            case messageDelete(MessageDelete)
            case messageDeleteBulk(MessageDeleteBulk)
            
            case messageReactionAdd(MessageReactionAdd)
            case messageReactionRemove(MessageReactionRemove)
            case messageReactionRemoveAll(MessageReactionRemoveAll)
            case messageReactionRemoveEmoji(MessageReactionRemoveEmoji)
            
            case presenceUpdate(PresenceUpdate)
            case requestPresenceUpdate(Identify.Presence)
            
            case stageInstanceCreate(StageInstance)
            case stageInstanceDelete(StageInstance)
            case stageInstanceUpdate(StageInstance)
            
            case typingStart(TypingStart)
            
            case userUpdate(DiscordUser)
            
            case voiceStateUpdate(VoiceState)
            case requestVoiceStateUpdate(VoiceStateUpdate)
            
            case voiceServerUpdate(VoiceServerUpdate)
            
            case webhooksUpdate(WebhooksUpdate)
            
            case applicationCommandPermissionsUpdate(GuildApplicationCommandPermissions)
            
            case autoModerationRuleCreate(AutoModerationRule)
            case autoModerationRuleUpdate(AutoModerationRule)
            case autoModerationRuleDelete(AutoModerationRule)
            
            case autoModerationActionExecution(AutoModerationActionExecution)
            
            public var correspondingIntents: [Intent] {
                switch self {
                case .heartbeat, .identify, .hello, .ready, .resume, .resumed, .invalidSession, .requestGuildMembers, .requestPresenceUpdate, .requestVoiceStateUpdate, .interactionCreate, .applicationCommandPermissionsUpdate, .userUpdate, .voiceServerUpdate:
                    return []
                case .guildCreate, .guildUpdate, .guildDelete, .guildMembersChunk, .guildRoleCreate, .guildRoleUpdate, .guildRoleDelete, .channelCreate, .channelUpdate, .channelDelete, .threadCreate, .threadUpdate, .threadDelete, .threadSyncList, .threadMemberUpdate, .stageInstanceCreate, .stageInstanceDelete, .stageInstanceUpdate:
                    return [.guilds]
                case .channelPinsUpdate:
                    return [.guilds, .directMessages]
                case .threadMembersUpdate, .guildMemberAdd, .guildMemberRemove, .guildMemberUpdate:
                    return [.guilds, .guildMembers]
                case .guildAuditLogEntryCreate, .guildBanAdd, .guildBanRemove:
                    return [.guildModeration]
                case .guildEmojisUpdate, .guildStickersUpdate:
                    return [.guildEmojisAndStickers]
                case .guildIntegrationsUpdate, .integrationCreate, .integrationUpdate, .integrationDelete:
                    return [.guildIntegrations]
                case .webhooksUpdate:
                    return [.guildWebhooks]
                case .inviteCreate, .inviteDelete:
                    return [.guildInvites]
                case .voiceStateUpdate:
                    return [.guildVoiceStates]
                case .presenceUpdate:
                    return [.guildPresences]
                case .messageCreate, .messageUpdate, .messageDelete:
                    return [.guildMessages, .directMessages]
                case .messageDeleteBulk:
                    return [.guildMessages]
                case .messageReactionAdd, .messageReactionRemove, .messageReactionRemoveAll, .messageReactionRemoveEmoji:
                    return [.guildMessageReactions]
                case .typingStart:
                    return [.guildMessageTyping]
                case .guildScheduledEventCreate, .guildScheduledEventUpdate, .guildScheduledEventDelete, .guildScheduledEventUserAdd, .guildScheduledEventUserRemove:
                    return [.guildScheduledEvents]
                case .autoModerationRuleCreate, .autoModerationRuleUpdate, .autoModerationRuleDelete:
                    return [.autoModerationConfiguration]
                case .autoModerationActionExecution:
                    return [.autoModerationExecution]
                }
            }
        }
        
        /// Read `helpAnchor` for help about each error case.
        public enum GatewayDecodingError: LocalizedError {
            case unhandledDispatchEvent(type: String?)
            
            public var errorDescription: String? {
                switch self {
                case let .unhandledDispatchEvent(type):
                    return "unhandledDispatchEvent(type: \(type ?? "nil"))"
                }
            }
            
            public var helpAnchor: String? {
                switch self {
                case let .unhandledDispatchEvent(type):
                    return "The dispatch event type '\(type ?? "nil")' is unhandled. This is probably a new Discord event which is not yet officially documented. I actively look for new events, and check Discord docs, so there is nothing to worry about. The library will support this event when it should"
                }
            }
        }
        
        enum CodingKeys: String, CodingKey {
            case opcode = "op"
            case data = "d"
            case sequenceNumber = "s"
            case type = "t"
        }
        
        public var opcode: Opcode
        public var data: Payload?
        public var sequenceNumber: Int?
        public var type: String?
        
        public init(
            opcode: Opcode,
            data: Payload? = nil,
            sequenceNumber: Int? = nil,
            type: String? = nil
        ) {
            self.opcode = opcode
            self.data = data
            self.sequenceNumber = sequenceNumber
            self.type = type
        }
        
        public init(from decoder: Decoder) throws {
            let container = try decoder.container(keyedBy: CodingKeys.self)
            self.opcode = try container.decode(Opcode.self, forKey: .opcode)
            self.sequenceNumber = try container.decodeIfPresent(Int.self, forKey: .sequenceNumber)
            self.type = try container.decodeIfPresent(String.self, forKey: .type)
            
            func decodeData<D: Decodable>(as type: D.Type = D.self) throws -> D {
                try container.decode(D.self, forKey: .data)
            }
            
            switch opcode {
            case .heartbeat, .heartbeatAccepted, .reconnect:
                guard try container.decodeNil(forKey: .data) else {
                    throw DecodingError.typeMismatch(Optional<Never>.self, .init(
                        codingPath: container.codingPath,
                        debugDescription: "`\(opcode)` opcode is supposed to have no data."
                    ))
                }
                self.data = nil
            case .identify, .presenceUpdate, .voiceStateUpdate, .resume, .requestGuildMembers:
                throw DecodingError.dataCorrupted(.init(
                    codingPath: container.codingPath,
                    debugDescription: "'\(opcode)' opcode is supposed to never be received."
                ))
            case .invalidSession:
                self.data = try .invalidSession(canResume: decodeData())
            case .hello:
                self.data = try .hello(decodeData())
            case .dispatch:
                switch self.type {
                case "READY":
                    self.data = try .ready(decodeData())
                case "RESUMED":
                    self.data = .resumed
                case "CHANNEL_CREATE":
                    self.data = try .channelCreate(decodeData())
                case "CHANNEL_UPDATE":
                    self.data = try .channelUpdate(decodeData())
                case "CHANNEL_DELETE":
                    self.data = try .channelDelete(decodeData())
                case "CHANNEL_PINS_UPDATE":
                    self.data = try .channelPinsUpdate(decodeData())
                case "THREAD_CREATE":
                    self.data = try .threadCreate(decodeData())
                case "THREAD_UPDATE":
                    self.data = try .threadUpdate(decodeData())
                case "THREAD_DELETE":
                    self.data = try .threadDelete(decodeData())
                case "THREAD_LIST_SYNC":
                    self.data = try .threadSyncList(decodeData())
                case "THREAD_MEMBER_UPDATE":
                    self.data = try .threadMemberUpdate(decodeData())
                case "THREAD_MEMBERS_UPDATE":
                    self.data = try .threadMembersUpdate(decodeData())
                case "GUILD_CREATE":
                    self.data = try .guildCreate(decodeData())
                case "GUILD_UPDATE":
                    self.data = try .guildUpdate(decodeData())
                case "GUILD_DELETE":
                    self.data = try .guildDelete(decodeData())
                case "GUILD_BAN_ADD":
                    self.data = try .guildBanAdd(decodeData())
                case "GUILD_BAN_REMOVE":
                    self.data = try .guildBanRemove(decodeData())
                case "GUILD_EMOJIS_UPDATE":
                    self.data = try .guildEmojisUpdate(decodeData())
                case "GUILD_STICKERS_UPDATE":
                    self.data = try .guildStickersUpdate(decodeData())
                case "GUILD_INTEGRATIONS_UPDATE":
                    self.data = try .guildIntegrationsUpdate(decodeData())
                case "GUILD_MEMBER_ADD":
                    self.data = try .guildMemberAdd(decodeData())
                case "GUILD_MEMBER_REMOVE":
                    self.data = try .guildMemberRemove(decodeData())
                case "GUILD_MEMBER_UPDATE":
                    self.data = try .guildMemberUpdate(decodeData())
                case "GUILD_MEMBERS_CHUNK":
                    self.data = try .guildMembersChunk(decodeData())
                case "GUILD_ROLE_CREATE":
                    self.data = try .guildRoleCreate(decodeData())
                case "GUILD_ROLE_UPDATE":
                    self.data = try .guildRoleUpdate(decodeData())
                case "GUILD_ROLE_DELETE":
                    self.data = try .guildRoleDelete(decodeData())
                case "GUILD_SCHEDULED_EVENT_CREATE":
                    self.data = try .guildScheduledEventCreate(decodeData())
                case "GUILD_SCHEDULED_EVENT_UPDATE":
                    self.data = try .guildScheduledEventUpdate(decodeData())
                case "GUILD_SCHEDULED_EVENT_DELETE":
                    self.data = try .guildScheduledEventDelete(decodeData())
                case "GUILD_SCHEDULED_EVENT_USER_ADD":
                    self.data = try .guildScheduledEventUserAdd(decodeData())
                case "GUILD_SCHEDULED_EVENT_USER_REMOVE":
                    self.data = try .guildScheduledEventUserRemove(decodeData())
                case "GUILD_AUDIT_LOG_ENTRY_CREATE":
                    self.data = try .guildAuditLogEntryCreate(decodeData())
                case "INTEGRATION_CREATE":
                    self.data = try .integrationCreate(decodeData())
                case "INTEGRATION_UPDATE":
                    self.data = try .integrationUpdate(decodeData())
                case "INTEGRATION_DELETE":
                    self.data = try .integrationDelete(decodeData())
                case "INTERACTION_CREATE":
                    self.data = try .interactionCreate(decodeData())
                case "INVITE_CREATE":
                    self.data = try .inviteCreate(decodeData())
                case "INVITE_DELETE":
                    self.data = try .inviteDelete(decodeData())
                case "MESSAGE_CREATE":
                    self.data = try .messageCreate(decodeData())
                case "MESSAGE_UPDATE":
                    self.data = try .messageUpdate(decodeData())
                case "MESSAGE_DELETE":
                    self.data = try .messageDelete(decodeData())
                case "MESSAGE_DELETE_BULK":
                    self.data = try .messageDeleteBulk(decodeData())
                case "MESSAGE_REACTION_ADD":
                    self.data = try .messageReactionAdd(decodeData())
                case "MESSAGE_REACTION_REMOVE":
                    self.data = try .messageReactionRemove(decodeData())
                case "MESSAGE_REACTION_REMOVE_ALL":
                    self.data = try .messageReactionRemoveAll(decodeData())
                case "MESSAGE_REACTION_REMOVE_EMOJI":
                    self.data = try .messageReactionRemoveEmoji(decodeData())
                case "PRESENCE_UPDATE":
                    self.data = try .presenceUpdate(decodeData())
                case "STAGE_INSTANCE_CREATE":
                    self.data = try .stageInstanceCreate(decodeData())
                case "STAGE_INSTANCE_DELETE":
                    self.data = try .stageInstanceDelete(decodeData())
                case "STAGE_INSTANCE_UPDATE":
                    self.data = try .stageInstanceUpdate(decodeData())
                case "TYPING_START":
                    self.data = try .typingStart(decodeData())
                case "USER_UPDATE":
                    self.data = try .userUpdate(decodeData())
                case "VOICE_STATE_UPDATE":
                    self.data = try .voiceStateUpdate(decodeData())
                case "VOICE_SERVER_UPDATE":
                    self.data = try .voiceServerUpdate(decodeData())
                case "WEBHOOKS_UPDATE":
                    self.data = try .webhooksUpdate(decodeData())
                case "APPLICATION_COMMAND_PERMISSIONS_UPDATE":
                    self.data = try .applicationCommandPermissionsUpdate(decodeData())
                case "AUTO_MODERATION_RULE_CREATE":
                    self.data = try .autoModerationRuleCreate(decodeData())
                case "AUTO_MODERATION_RULE_UPDATE":
                    self.data = try .autoModerationRuleUpdate(decodeData())
                case "AUTO_MODERATION_RULE_DELETE":
                    self.data = try .autoModerationRuleDelete(decodeData())
                case "AUTO_MODERATION_ACTION_EXECUTION":
                    self.data = try .autoModerationActionExecution(decodeData())
                default:
                    throw GatewayDecodingError.unhandledDispatchEvent(type: self.type)
                }
            }
        }
        
        /// Read `helpAnchor` for help about each error case.
        enum EncodingError: LocalizedError {
            case notSupposedToBeSent(message: String)
            
            public var errorDescription: String? {
                switch self {
                case let .notSupposedToBeSent(message):
                    return "notSupposedToBeSent(\(message))"
                }
            }
            
            public var helpAnchor: String? {
                switch self {
                case let .notSupposedToBeSent(message):
                    return "This event is not supposed to be sent at all. This could be a library issue, please report at https://github.com/MahdiBM/DiscordBM/issues. Message: \(message)"
                }
            }
        }
        
        public func encode(to encoder: Encoder) throws {
            var container = encoder.container(keyedBy: CodingKeys.self)
            
            switch self.opcode {
            case .dispatch, .reconnect, .invalidSession, .heartbeatAccepted, .hello:
                throw EncodingError.notSupposedToBeSent(
                    message: "`\(self.opcode.rawValue)` opcode is supposed to never be sent."
                )
            default: break
            }
            try container.encode(self.opcode, forKey: .opcode)
            
            if self.sequenceNumber != nil {
                throw EncodingError.notSupposedToBeSent(
                    message: "'sequenceNumber' is supposed to never be sent but wasn't nil (\(String(describing: sequenceNumber))."
                )
            }
            if self.type != nil {
                throw EncodingError.notSupposedToBeSent(
                    message: "'type' is supposed to never be sent but wasn't nil (\(String(describing: type))."
                )
            }
            
            switch self.data {
            case .none:
                try container.encodeNil(forKey: .data)
            case let .heartbeat(lastSequenceNumber):
                try container.encode(lastSequenceNumber, forKey: .data)
            case let .identify(payload):
                try container.encode(payload, forKey: .data)
            case let .resume(payload):
                try container.encode(payload, forKey: .data)
            case let .requestGuildMembers(payload):
                try container.encode(payload, forKey: .data)
            case let .requestPresenceUpdate(payload):
                try container.encode(payload, forKey: .data)
            case let .requestVoiceStateUpdate(payload):
                try container.encode(payload, forKey: .data)
            default:
                throw EncodingError.notSupposedToBeSent(
                    message: "'\(self)' data is supposed to never be sent."
                )
            }
        }
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#identify
    public struct Identify: Sendable, Codable {
        
        /// https://discord.com/developers/docs/topics/gateway-events#identify-identify-connection-properties
        public struct ConnectionProperties: Sendable, Codable {
            public let os: String = {
#if os(macOS)
                return "macOS"
#elseif os(Linux)
                return "Linux"
#elseif os(iOS)
                return "iOS"
#elseif os(watchOS)
                return "watchOS"
#elseif os(tvOS)
                return "tvOS"
#elseif os(Windows)
                return "Windows"
#elseif os(Android)
                return "Android"
#else
                return "Unknown-OS"
#endif
            }()
            public let browser = "DiscordBM"
            public let device = "DiscordBM"
            
            enum CodingKeys: String, CodingKey {
                case os = "$os"
                case browser = "$browser"
                case device = "$device"
            }
        }
        
        /// https://discord.com/developers/docs/topics/gateway-events#update-presence-gateway-presence-update-structure
        public struct Presence: Sendable, Codable {
            public var since: Int?
            public var activities: [Activity]
            public var status: Status
            public var afk: Bool
            
            public init(since: Int? = nil, activities: [Activity], status: Status, afk: Bool) {
                self.since = since
                self.activities = activities
                self.status = status
                self.afk = afk
            }
        }
        
        public var token: Secret
        public var properties = ConnectionProperties()
        /// DiscordBM supports the better `Transport Compression`, but not `Payload Compression`.
        /// Setting this to `true` will only cause problems.
        /// Instead, use the `compression` parameter in the `BotGatewayManager` initializer.
        public var compress: Bool?
        public var large_threshold: Int?
        public var shard: IntPair?
        public var presence: Presence?
        public var intents: IntBitField<Intent>
        
        public init(token: Secret, large_threshold: Int? = nil, shard: IntPair? = nil, presence: Presence? = nil, intents: [Intent]) {
            self.token = token
            self.large_threshold = large_threshold
            self.shard = shard
            self.presence = presence
            self.intents = .init(intents)
        }
        
        public init(token: String, large_threshold: Int? = nil, shard: IntPair? = nil, presence: Presence? = nil, intents: [Intent]) {
            self.token = Secret(token)
            self.large_threshold = large_threshold
            self.shard = shard
            self.presence = presence
            self.intents = .init(intents)
        }
    }
    
    /// https://discord.com/developers/docs/topics/gateway#gateway-intents
    public enum Intent: Int, Sendable, Codable, CaseIterable {
        case guilds = 0
        case guildMembers = 1
        case guildModeration = 2
        case guildEmojisAndStickers = 3
        case guildIntegrations = 4
        case guildWebhooks = 5
        case guildInvites = 6
        case guildVoiceStates = 7
        case guildPresences = 8
        case guildMessages = 9
        case guildMessageReactions = 10
        case guildMessageTyping = 11
        case directMessages = 12
        case directMessageReactions = 13
        case directMessageTyping = 14
        case messageContent = 15
        case guildScheduledEvents = 16
        case autoModerationConfiguration = 20
        case autoModerationExecution = 21
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#resume-resume-structure
    public struct Resume: Sendable, Codable {
        public var token: Secret
        public var session_id: String
        public var seq: Int
        
        public init(token: Secret, session_id: String, sequence: Int) {
            self.token = token
            self.session_id = session_id
            self.seq = sequence
        }
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#update-presence-status-types
    public enum Status: String, Sendable, Codable, ToleratesStringDecodeMarker {
        case online = "online"
        case doNotDisturb = "dnd"
        case afk = "idle"
        case offline = "offline"
        case invisible = "invisible"
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#hello-hello-structure
    public struct Hello: Sendable, Codable {
        public var heartbeat_interval: Int
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#ready-ready-event-fields
    public struct Ready: Sendable, Codable {
        public var v: Int
        public var user: DiscordUser
        public var guilds: [UnavailableGuild]
        public var session_id: String
        public var resume_gateway_url: String?
        public var shard: IntPair?
        public var application: PartialApplication
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#thread-delete
    public struct ThreadDelete: Sendable, Codable {
        public var id: String
        public var type: DiscordChannel.Kind
        public var guild_id: String?
        public var parent_id: String?
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#thread-list-sync-thread-list-sync-event-fields
    public struct ThreadListSync: Sendable, Codable {
        public var guild_id: String
        public var channel_ids: [String]?
        public var threads: [DiscordChannel]
        public var members: [ThreadMember]
    }
    
    /// A ``ThreadMember`` with a `guild_id` field.
    /// https://discord.com/developers/docs/topics/gateway-events#thread-member-update
    public struct ThreadMemberUpdate: Sendable, Codable {
        public var id: String
        public var user_id: String?
        public var join_timestamp: DiscordTimestamp
        /// FIXME:
        /// The field is documented but doesn't say what exactly it is.
        /// Discord says: "any user-thread settings, currently only used for notifications".
        /// I think currently it's set to `1` or `0` depending on if you have notifications
        /// enabled for the thread?
        public var flags: Int
        public var guild_id: String
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#thread-members-update-thread-members-update-event-fields
    public struct ThreadMembersUpdate: Sendable, Codable {
        
        /// A ``ThreadMember`` with some extra fields.
        /// https://discord.com/developers/docs/resources/channel#thread-member-object-thread-member-structure
        /// https://discord.com/developers/docs/topics/gateway-events#thread-members-update-thread-members-update-event-fields
        public struct ThreadMember: Sendable, Codable {
            
            /// A ``PresenceUpdate`` with nullable `guild_id`.
            /// https://discord.com/developers/docs/topics/gateway-events#presence-update-presence-update-event-fields
            public struct ThreadMemberPresenceUpdate: Sendable, Codable {
                public var user: PartialUser
                public var guild_id: String?
                public var status: Status
                public var activities: [Activity]
                public var client_status: ClientStatus
            }
            
            public var id: String
            public var user_id: String?
            public var join_timestamp: DiscordTimestamp
            /// FIXME:
            /// The field is documented but doesn't say what exactly it is.
            /// Discord says: "any user-thread settings, currently only used for notifications".
            /// I think currently it's set to `1` or `0` depending on if you have notifications
            /// enabled for the thread?
            public var flags: Int
            public var member: Guild.Member
            public var presence: ThreadMemberPresenceUpdate?
        }
        
        public var id: String
        public var guild_id: String
        public var member_count: Int
        public var added_members: [ThreadMember]?
        public var removed_member_ids: [String]?
    }
    
    /// A `Guild` object with extra fields.
    /// https://discord.com/developers/docs/resources/guild#guild-object-guild-structure
    /// https://discord.com/developers/docs/topics/gateway-events#guild-create-guild-create-extra-fields
    public struct GuildCreate: Sendable, Codable {
        public var id: String
        public var name: String
        public var icon: String?
        public var icon_hash: String?
        public var splash: String?
        public var discovery_splash: String?
        public var owner: Bool?
        public var owner_id: String
        public var permissions: StringBitField<Permission>?
        /// Deprecated
        public var region: String?
        public var afk_channel_id: String?
        public var afk_timeout: Int
        public var widget_enabled: Bool?
        public var widget_channel_id: String?
        public var verification_level: Guild.VerificationLevel
        public var default_message_notifications: Guild.DefaultMessageNotificationLevel
        public var explicit_content_filter: Guild.ExplicitContentFilterLevel
        public var roles: [Role]
        public var emojis: [PartialEmoji]
        public var features: [Guild.Feature]
        public var mfa_level: Guild.MFALevel
        public var application_id: String?
        public var system_channel_id: String?
        public var system_channel_flags: IntBitField<Guild.SystemChannelFlag>
        public var rules_channel_id: String?
        public var safety_alerts_channel_id: String?
        public var max_presences: Int?
        public var max_members: Int?
        public var vanity_url_code: String?
        public var description: String?
        public var banner: String?
        public var premium_tier: Guild.PremiumTier
        public var premium_subscription_count: Int?
        public var preferred_locale: DiscordLocale
        public var public_updates_channel_id: String?
        public var max_video_channel_users: Int?
        public var max_stage_video_channel_users: Int?
        public var approximate_member_count: Int?
        public var approximate_presence_count: Int?
        public var welcome_screen: [Guild.WelcomeScreen]?
        public var nsfw_level: Guild.NSFWLevel
        public var stickers: [Sticker]?
        public var premium_progress_bar_enabled: Bool
        public var `lazy`: Bool?
        public var hub_type: String?
        public var nsfw: Bool
        public var application_command_counts: [String: Int]?
        public var embedded_activities: [Gateway.Activity]?
        public var version: Int?
        public var guild_id: String?
        /// Extra fields:
        public var joined_at: DiscordTimestamp
        public var large: Bool
        public var unavailable: Bool?
        public var member_count: Int
        public var voice_states: [PartialVoiceState]
        public var members: [Guild.Member]
        public var channels: [DiscordChannel]
        public var threads: [DiscordChannel]
        public var presences: [Gateway.PartialPresenceUpdate]
        public var guild_scheduled_events: [GuildScheduledEvent]
        public var stage_instances: [StageInstance]
        
        public mutating func update(with new: Guild) {
            self.id = new.id
            self.name = new.name
            self.icon = new.icon
            self.icon_hash = new.icon_hash
            self.splash = new.splash
            self.discovery_splash = new.discovery_splash
            self.owner = new.owner
            self.owner_id = new.owner_id
            self.permissions = new.permissions
            self.region = new.region
            self.afk_channel_id = new.afk_channel_id
            self.afk_timeout = new.afk_timeout
            self.widget_enabled = new.widget_enabled
            self.widget_channel_id = new.widget_channel_id
            self.verification_level = new.verification_level
            self.default_message_notifications = new.default_message_notifications
            self.explicit_content_filter = new.explicit_content_filter
            self.roles = new.roles
            self.emojis = new.emojis
            self.features = new.features
            self.mfa_level = new.mfa_level
            self.application_id = new.application_id
            self.system_channel_id = new.system_channel_id
            self.system_channel_flags = new.system_channel_flags
            self.rules_channel_id = new.rules_channel_id
            self.max_presences = new.max_presences
            self.max_members = new.max_members
            self.vanity_url_code = new.vanity_url_code
            self.description = new.description
            self.banner = new.banner
            self.premium_tier = new.premium_tier
            self.premium_subscription_count = new.premium_subscription_count
            self.preferred_locale = new.preferred_locale
            self.public_updates_channel_id = new.public_updates_channel_id
            self.max_video_channel_users = new.max_video_channel_users
            self.max_stage_video_channel_users = new.max_stage_video_channel_users
            self.approximate_member_count = new.approximate_member_count
            self.approximate_presence_count = new.approximate_presence_count
            self.welcome_screen = new.welcome_screen
            self.nsfw_level = new.nsfw_level
            self.stickers = new.stickers
            self.premium_progress_bar_enabled = new.premium_progress_bar_enabled
            self.`lazy` = new.`lazy`
            self.hub_type = new.hub_type
            self.nsfw = new.nsfw
            self.application_command_counts = new.application_command_counts
            self.embedded_activities = new.embedded_activities
            self.version = new.version
            self.guild_id = new.guild_id
        }
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#channel-pins-update-channel-pins-update-event-fields
    public struct ChannelPinsUpdate: Sendable, Codable {
        public var guild_id: String?
        public var channel_id: String
        public var last_pin_timestamp: DiscordTimestamp?
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#guild-ban-add-guild-ban-add-event-fields
    public struct GuildBan: Sendable, Codable {
        public var guild_id: String
        public var user: DiscordUser
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#guild-emojis-update-guild-emojis-update-event-fields
    public struct GuildEmojisUpdate: Sendable, Codable {
        public var guild_id: String
        public var emojis: [PartialEmoji]
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#guild-stickers-update-guild-stickers-update-event-fields
    public struct GuildStickersUpdate: Sendable, Codable {
        public var guild_id: String
        public var stickers: [Sticker]
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#guild-integrations-update-guild-integrations-update-event-fields
    public struct GuildIntegrationsUpdate: Sendable, Codable {
        public var guild_id: String
    }
    
    /// A ``Guild.Member`` with an extra `guild_id` field.
    /// https://discord.com/developers/docs/resources/guild#guild-member-object
    public struct GuildMemberAdd: Sendable, Codable {
        public var guild_id: String
        public var roles: [String]
        public var hoisted_role: String?
        public var user: DiscordUser
        public var nick: String?
        public var avatar: String?
        public var joined_at: DiscordTimestamp
        public var premium_since: DiscordTimestamp?
        public var deaf: Bool?
        public var mute: Bool?
        public var pending: Bool?
        public var is_pending: Bool?
        public var flags: IntBitField<Guild.Member.Flag>?
        public var permissions: StringBitField<Permission>?
        public var communication_disabled_until: DiscordTimestamp?
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#guild-member-remove-guild-member-remove-event-fields
    public struct GuildMemberRemove: Sendable, Codable {
        public var guild_id: String
        public var user: DiscordUser
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#guild-members-chunk
    public struct GuildMembersChunk: Sendable, Codable {
        public var guild_id: String
        public var members: [Guild.Member]
        public var chunk_index: Int
        public var chunk_count: Int
        public var not_found: [String]?
        public var presences: [PartialPresenceUpdate]?
        public var nonce: String?
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#request-guild-members
    public struct RequestGuildMembers: Sendable, Codable {
        public var guild_id: String
        public var query: String = ""
        public var limit: Int = 0
        public var presences: Bool?
        public var user_ids: [String]?
        public var nonce: String?
        
        public init(guild_id: String, query: String = "", limit: Int = 0, presences: Bool? = nil, user_ids: [String]? = nil, nonce: String? = nil) {
            self.guild_id = guild_id
            self.query = query
            self.limit = limit
            self.presences = presences
            self.user_ids = user_ids
            self.nonce = nonce
        }
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#guild-role-create-guild-role-create-event-fields
    public struct GuildRole: Sendable, Codable {
        public var guild_id: String
        public var role: Role
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#guild-role-delete
    public struct GuildRoleDelete: Sendable, Codable {
        public var guild_id: String
        public var role_id: String
        public var version: Int?
    }
    
    /// Not the same as what Discord calls `Guild Scheduled Event User`.
    /// This is used for guild-scheduled-event-user add and remove events.
    /// https://discord.com/developers/docs/topics/gateway-events#guild-scheduled-event-user-add-guild-scheduled-event-user-add-event-fields
    public struct GuildScheduledEventUser: Sendable, Codable {
        public var guild_scheduled_event_id: String
        public var user_id: String
        public var guild_id: String
    }
    
    /// An ``Integration`` with an extra `guild_id` field.
    /// https://discord.com/developers/docs/topics/gateway-events#integration-create
    /// https://discord.com/developers/docs/resources/guild#integration-object
    public struct IntegrationCreate: Sendable, Codable {
        public var id: String
        public var name: String
        public var type: Integration.Kind
        public var enabled: Bool
        public var syncing: Bool?
        public var role_id: String?
        public var enable_emoticons: Bool?
        public var expire_behavior: Integration.ExpireBehavior?
        public var expire_grace_period: Int?
        public var user: DiscordUser?
        public var account: IntegrationAccount
        public var synced_at: DiscordTimestamp?
        public var subscriber_count: Int?
        public var revoked: Bool?
        public var application: IntegrationApplication?
        public var guild_id: String
        public var scopes: [OAuth2Scope]?
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#integration-delete-integration-delete-event-fields
    public struct IntegrationDelete: Sendable, Codable {
        public var id: String
        public var guild_id: String
        public var application_id: String?
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#invite-create-invite-create-event-fields
    public struct InviteCreate: Sendable, Codable {
        
        /// https://discord.com/developers/docs/resources/invite#invite-object-invite-target-types
        public enum TargetKind: Int, Sendable, Codable, ToleratesIntDecodeMarker {
            case stream = 1
            case embeddedApplication = 2
        }
        
        public var channel_id: String
        public var code: String
        public var created_at: DiscordTimestamp
        public var guild_id: String?
        public var inviter: DiscordUser?
        public var max_age: Int
        public var max_uses: Int
        public var target_type: TargetKind?
        public var target_user: DiscordUser?
        public var target_application: PartialApplication?
        public var temporary: Bool
        public var uses: Int
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#invite-delete
    public struct InviteDelete: Sendable, Codable {
        public var channel_id: String
        public var guild_id: String?
        public var code: String
    }
    
    /// A ``Message`` object with a few extra fields.
    /// https://discord.com/developers/docs/topics/gateway-events#message-create
    /// https://discord.com/developers/docs/resources/channel#message-object
    public struct MessageCreate: Sendable, Codable {
        public var id: String
        public var channel_id: String
        public var author: PartialUser?
        public var content: String
        public var timestamp: DiscordTimestamp
        public var edited_timestamp: DiscordTimestamp?
        public var tts: Bool
        public var mention_everyone: Bool
        public var mention_roles: [String]
        public var mention_channels: [DiscordChannel.Message.ChannelMention]?
        public var attachments: [DiscordChannel.Message.Attachment]
        public var embeds: [Embed]
        public var reactions: [DiscordChannel.Message.Reaction]?
        public var nonce: StringOrInt?
        public var pinned: Bool
        public var webhook_id: String?
        public var type: DiscordChannel.Message.Kind
        public var activity: DiscordChannel.Message.Activity?
        public var application: PartialApplication?
        public var application_id: String?
        public var message_reference: DiscordChannel.Message.MessageReference?
        public var flags: IntBitField<DiscordChannel.Message.Flag>?
        public var referenced_message: DereferenceBox<MessageCreate>?
        public var interaction: MessageInteraction?
        public var thread: DiscordChannel?
        public var components: [Interaction.ActionRow]?
        public var sticker_items: [StickerItem]?
        public var stickers: [Sticker]?
        public var position: Int?
        public var role_subscription_data: RoleSubscriptionData?
        /// The extra fields:
        public var guild_id: String?
        public var member: Guild.PartialMember?
        public var mentions: [DiscordChannel.Message.MentionUser]
        
        public mutating func update(with partialMessage: DiscordChannel.PartialMessage) {
            self.id = partialMessage.id
            self.channel_id = partialMessage.channel_id
            self.author = partialMessage.author
            if let content = partialMessage.content {
                self.content = content
            }
            if let timestamp = partialMessage.timestamp {
                self.timestamp = timestamp
            }
            self.edited_timestamp = partialMessage.edited_timestamp
            if let tts = partialMessage.tts {
                self.tts = tts
            }
            if let mention_everyone = partialMessage.mention_everyone {
                self.mention_everyone = mention_everyone
            }
            if let mentions = partialMessage.mentions {
                self.mentions = mentions
            }
            if let mention_roles = partialMessage.mention_roles {
                self.mention_roles = mention_roles
            }
            self.mention_channels = partialMessage.mention_channels
            if let attachments = partialMessage.attachments {
                self.attachments = attachments
            }
            if let embeds = partialMessage.embeds {
                self.embeds = embeds
            }
            self.reactions = partialMessage.reactions
            self.nonce = partialMessage.nonce
            if let pinned = partialMessage.pinned {
                self.pinned = pinned
            }
            self.webhook_id = partialMessage.webhook_id
            if let type = partialMessage.type {
                self.type = type
            }
            if let activity = partialMessage.activity {
                self.activity = activity
            }
            self.application = partialMessage.application
            self.application_id = partialMessage.application_id
            self.message_reference = partialMessage.message_reference
            self.flags = partialMessage.flags
            if let referenced_message = partialMessage.referenced_message,
               var value = self.referenced_message?.value {
                value.update(with: referenced_message.value)
                self.referenced_message = .init(value: value)
            }
            self.interaction = partialMessage.interaction
            self.thread = partialMessage.thread
            self.components = partialMessage.components
            self.sticker_items = partialMessage.sticker_items
            self.stickers = partialMessage.stickers
            self.position = partialMessage.position
            self.role_subscription_data = partialMessage.role_subscription_data
            if let member = partialMessage.member {
                self.member = member
            }
            if let guildId = partialMessage.guild_id {
                self.guild_id = guildId
            }
        }
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#message-delete
    public struct MessageDelete: Sendable, Codable {
        public var id: String
        public var channel_id: String
        public var guild_id: String?
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#message-delete-bulk-message-delete-bulk-event-fields
    public struct MessageDeleteBulk : Sendable, Codable {
        public var ids: [String]
        public var channel_id: String
        public var guild_id: String?
    }
    
    public enum ReactionKind: Int, Sendable, Codable {
        case normal = 0
        case `super` = 1
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#message-reaction-add-message-reaction-add-event-fields
    public struct MessageReactionAdd: Sendable, Codable {
        public var type: ReactionKind
        public var user_id: String
        public var channel_id: String
        public var message_id: String
        public var guild_id: String?
        public var burst: Bool?
        public var member: Guild.Member?
        public var emoji: PartialEmoji
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#message-reaction-remove
    public struct MessageReactionRemove: Sendable, Codable {
        public var type: ReactionKind
        public var user_id: String
        public var channel_id: String
        public var message_id: String
        public var guild_id: String?
        public var burst: Bool?
        public var emoji: PartialEmoji
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#message-reaction-remove-all
    public struct MessageReactionRemoveAll: Sendable, Codable {
        public var channel_id: String
        public var message_id: String
        public var guild_id: String?
        public var burst: Bool?
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#message-reaction-remove-emoji
    public struct MessageReactionRemoveEmoji: Sendable, Codable {
        public var channel_id: String
        public var guild_id: String?
        public var message_id: String
        public var burst: Bool?
        public var emoji: PartialEmoji
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#client-status-object
    public struct ClientStatus: Sendable, Codable {
        public var desktop: Status?
        public var mobile: Status?
        public var web: Status?
        public var embedded: Status?
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#presence-update-presence-update-event-fields
    public struct PresenceUpdate: Sendable, Codable {
        public var user: PartialUser
        public var guild_id: String
        public var status: Status
        public var activities: [Activity]
        public var client_status: ClientStatus
    }
    
    /// Partial ``PresenceUpdate`` object.
    /// https://discord.com/developers/docs/topics/gateway-events#presence-update-presence-update-event-fields
    public struct PartialPresenceUpdate: Sendable, Codable {
        public var user: PartialUser?
        public var guild_id: String?
        public var status: Status?
        public var activities: [Activity]?
        public var client_status: ClientStatus
        
        public mutating func update(with presenceUpdate: Gateway.PresenceUpdate) {
            self.guild_id = presenceUpdate.guild_id
            self.status = presenceUpdate.status
            self.activities = presenceUpdate.activities
            self.client_status = presenceUpdate.client_status
        }
        
        public init(presenceUpdate: Gateway.PresenceUpdate) {
            self.user = presenceUpdate.user
            self.guild_id = presenceUpdate.guild_id
            self.status = presenceUpdate.status
            self.activities = presenceUpdate.activities
            self.client_status = presenceUpdate.client_status
        }
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#activity-object
    public struct Activity: Sendable, Codable {
        
        /// https://discord.com/developers/docs/topics/gateway-events#activity-object-activity-types
        public enum Kind: Int, Sendable, Codable, ToleratesIntDecodeMarker {
            case game = 0
            case streaming = 1
            case listening = 2
            case watching = 3
            case custom = 4
            case competing = 5
        }
        
        /// https://discord.com/developers/docs/topics/gateway-events#activity-object-activity-timestamps
        public struct Timestamps: Sendable, Codable {
            public var start: Int?
            public var end: Int?
            
            public init(start: Int? = nil, end: Int? = nil) {
                self.start = start
                self.end = end
            }
        }
        
        /// https://discord.com/developers/docs/topics/gateway-events#activity-object-activity-emoji
        public struct ActivityEmoji: Sendable, Codable {
            public var name: String
            public var id: String?
            public var animated: Bool?
            
            public init(name: String, id: String? = nil, animated: Bool? = nil) {
                self.name = name
                self.id = id
                self.animated = animated
            }
        }
        
        /// https://discord.com/developers/docs/topics/gateway-events#activity-object-activity-party
        public struct Party: Sendable, Codable {
            public var id: String?
            public var size: IntPair?
            
            public init(id: String? = nil, size: IntPair? = nil) {
                self.id = id
                self.size = size
            }
        }
        
        /// https://discord.com/developers/docs/topics/gateway-events#activity-object-activity-assets
        public struct Assets: Sendable, Codable {
            public var large_image: String?
            public var large_text: String?
            public var small_image: String?
            public var small_text: String?
            
            public init(large_image: String? = nil, large_text: String? = nil, small_image: String? = nil, small_text: String? = nil) {
                self.large_image = large_image
                self.large_text = large_text
                self.small_image = small_image
                self.small_text = small_text
            }
        }
        
        /// https://discord.com/developers/docs/topics/gateway-events#activity-object-activity-secrets
        public struct Secrets: Sendable, Codable {
            public var join: String?
            public var spectate: String?
            public var match: String?
            
            public init(join: String? = nil, spectate: String? = nil, match: String? = nil) {
                self.join = join
                self.spectate = spectate
                self.match = match
            }
        }
        
        /// https://discord.com/developers/docs/topics/gateway-events#activity-object-activity-flags
        public enum Flag: Int, Sendable {
            case instance = 0
            case join = 1
            case spectate = 2
            case joinRequest = 3
            case sync = 4
            case play = 5
            case partyPrivacyFriends = 6
            case partyPrivacyVoiceChannel = 7
            case embedded = 8
        }
        
        /// https://discord.com/developers/docs/topics/gateway-events#activity-object-activity-buttons
        public struct Button: Sendable, Codable {
            public var label: String
            public var url: String
            
            public init(label: String, url: String) {
                self.label = label
                self.url = url
            }
            
            public init(from decoder: Decoder) throws {
                if let container = try? decoder.container(keyedBy: CodingKeys.self) {
                    self.label = try container.decode(String.self, forKey: .label)
                    self.url = try container.decode(String.self, forKey: .url)
                } else {
                    self.label = try decoder.singleValueContainer().decode(String.self)
                    self.url = ""
                }
            }
        }
        
        public var name: String?
        public var type: Kind?
        public var url: String?
        public var id: String?
        public var created_at: Int?
        public var timestamps: Timestamps?
        public var application_id: String?
        public var details: String?
        public var state: String?
        public var emoji: ActivityEmoji?
        public var party: Party?
        public var party_id: String?
        public var assets: Assets?
        public var secrets: Secrets?
        public var instance: Bool?
        public var flags: IntBitField<Flag>?
        public var buttons: [Button]?
        public var sync_id: String?
        public var session_id: String?
        public var platform: String?
        public var supported_platforms: [String]?
        
        /// Bots are only able to send `name`, `type`, and optionally `url`.
        public init(name: String, type: Kind, url: String? = nil) {
            self.name = name
            self.type = type
            self.url = url
        }
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#typing-start-typing-start-event-fields
    public struct TypingStart: Sendable, Codable {
        public var channel_id: String
        public var guild_id: String?
        public var user_id: String
        public var timestamp: Int
        public var member: Guild.Member?
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#voice-server-update-voice-server-update-event-fields
    public struct VoiceServerUpdate: Sendable, Codable {
        public var token: String
        public var guild_id: String
        public var endpoint: String?
    }
    
    /// https://discord.com/developers/docs/topics/gateway-events#webhooks-update-webhooks-update-event-fields
    public struct WebhooksUpdate: Sendable, Codable {
        public var guild_id: String
        public var channel_id: String
    }
    
    /// https://discord.com/developers/docs/topics/gateway#get-gateway
    public struct URL: Sendable, Codable {
        public var url: String
    }
    
    /// https://discord.com/developers/docs/topics/gateway#get-gateway-bot-json-response
    public struct BotConnectionInfo: Sendable, Codable {
        
        /// https://discord.com/developers/docs/topics/gateway#session-start-limit-object-session-start-limit-structure
        public struct SessionStartLimit: Sendable, Codable {
            public var total: Int
            public var remaining: Int
            public var reset_after: Int
            public var max_concurrency: Int
        }
        
        public var url: String
        public var shards: Int
        public var session_start_limit: SessionStartLimit
    }
}
